#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <Windows.h>
#include <oleacc.h>
#include <dwmapi.h>
#include <winuser.h>
#include "ktalk_napi.h"

#define FOREGROUND_TIMER_MS 83 // 12 fps

struct watch_window_target
{
  HWND hwnd;
  HWINEVENTHOOK location_hook;
  HWINEVENTHOOK destroy_hook;
  HWINEVENTHOOK foreground_hook;
  HWINEVENTHOOK minimized_hook;
  bool is_focused;
  UINT_PTR timer_id;
};

static UINT WM_OVERLAY_UIPI_TEST = WM_NULL;

static struct watch_window_target watch_target_info = {
  .hwnd = NULL,
  .location_hook = NULL,
  .destroy_hook = NULL,
  .foreground_hook = NULL,
  .minimized_hook = NULL,
  .is_focused = false,
  .timer_id = NULL
};

static HWND foreground_window = NULL;

static VOID CALLBACK hook_proc(HWINEVENTHOOK, DWORD, HWND, LONG, LONG, DWORD, DWORD);

static bool has_uipi_access(HWND hwnd) {
  SetLastError(ERROR_SUCCESS);
  PostMessage(hwnd, WM_OVERLAY_UIPI_TEST, 0, 0);
  return GetLastError() != ERROR_ACCESS_DENIED;
}

static bool get_title(HWND hwnd, char** title) {
  SetLastError(0);
  int titleLength = GetWindowTextLengthW(hwnd);
  if (titleLength == 0) {
    if (GetLastError() != 0) {
      return false;
    }
    else {
      *title = NULL;
      return true;
    }
  }

  LPWSTR titleUtf16 = malloc(sizeof(WCHAR) * ((size_t)titleLength + 1));
  if (GetWindowTextW(hwnd, titleUtf16, titleLength + 1) == FALSE) {
    free(titleUtf16);
    return false;
  }
  int buffLenUtf8 = WideCharToMultiByte(CP_UTF8, 0, titleUtf16, -1, NULL, 0, NULL, NULL);
  if (buffLenUtf8 == FALSE) {
    free(titleUtf16);
    return false;
  }
  *title = malloc(buffLenUtf8);
  if (WideCharToMultiByte(CP_UTF8, 0, titleUtf16, -1, *title, buffLenUtf8, NULL, NULL) == FALSE) {
    free(titleUtf16);
    free(*title);
    return false;
  }
  return true;
}

static bool get_content_bounds(HWND hwnd, struct ow_window_bounds* bounds) {
  RECT rect;
  // if (GetWindowRect(hwnd, &rect) == FALSE) {
  //   return false;
  // }

  if (DwmGetWindowAttribute(hwnd, DWMWA_EXTENDED_FRAME_BOUNDS, &rect, sizeof(RECT)) != S_OK){
    return false;
  }

  bounds->x = rect.left;
  bounds->y = rect.top;
  bounds->width = rect.right - rect.left;
  bounds->height = rect.bottom - rect.top;
  return true;
}

static bool MSAA_check_window_focused_state(HWND hwnd) {
  HRESULT hr;
  IAccessible* pAcc = NULL;
  VARIANT varChildSelf;
  VariantInit(&varChildSelf);
  hr = AccessibleObjectFromEvent(hwnd, OBJID_WINDOW, CHILDID_SELF, &pAcc, &varChildSelf);
  if (hr != S_OK || pAcc == NULL) {
    VariantClear(&varChildSelf);
    return false;
  }
  VARIANT varState;
  VariantInit(&varState);
  hr = pAcc->lpVtbl->get_accState(pAcc, varChildSelf, &varState);

  bool is_focused = false;
  if (hr == S_OK && varState.vt == VT_I4) {
    is_focused = (varState.lVal & STATE_SYSTEM_FOCUSED);
  }
  VariantClear(&varState);
  VariantClear(&varChildSelf);
  pAcc->lpVtbl->Release(pAcc);
  return is_focused;
}

static void handle_movesize_event(struct watch_window_target* target_info) {
  struct ow_window_bounds bounds;
  if (get_content_bounds(target_info->hwnd, &bounds)) {
    struct ow_event e = {
      .type = OW_MOVERESIZE,
      .data.moveresize = {
        .bounds = bounds
      }
    };
    watch_window_emit_event(&e);
  }
}

static void check_and_handle_window(HWND hwnd, struct watch_window_target* target_info) {
  // ignore fake ghost windows
  if (IsHungAppWindow(hwnd)) {
    return;
  }

  if (target_info->hwnd != hwnd) {
    if (target_info->is_focused) {
      target_info->is_focused = false;
      struct ow_event e = { .type = OW_BLUR };
      watch_window_emit_event(&e);
    }
  }
  else if (target_info->hwnd == hwnd) {
    if (!target_info->is_focused) {
      target_info->is_focused = true;
      struct ow_event e = { .type = OW_FOCUS };
      watch_window_emit_event(&e);
    }
    return;
  }

}

static void handle_new_foreground(HWND hwnd) {
  foreground_window = hwnd;
  check_and_handle_window(foreground_window, &watch_target_info);
}

static VOID CALLBACK hook_proc(
  HWINEVENTHOOK hWinEventHook, DWORD event, HWND hwnd, LONG idObject, LONG idChild,
  DWORD idEventThread, DWORD dwmsEventTime
) {
  /* char* e_str =
    event == EVENT_SYSTEM_FOREGROUND ? "SYS_FOREGROUND"
    : event == EVENT_SYSTEM_MINIMIZEEND ? "SYS_MINIMIZEEND"
    : event == EVENT_OBJECT_NAMECHANGE ? "OBJ_NAMECHANGE"
    : event == EVENT_OBJECT_LOCATIONCHANGE ? "OBJ_LOCATIONCHANGE"
    : event == EVENT_OBJECT_DESTROY ? "OBJ_DESTROY"
    : "(unknown)";
  printf("[%d] %s hwnd=%p idObject=%d idChild=%d\n", dwmsEventTime, e_str, hwnd, idObject, idChild); */

  // printf("[%d] hwnd=%p, target_hwnd=%p idObject=%d idChild=%d\n", dwmsEventTime, hwnd, target_info.hwnd, idObject, idChild);
  if (event == EVENT_OBJECT_DESTROY) {
    if (hwnd == watch_target_info.hwnd && idObject == OBJID_WINDOW && idChild == CHILDID_SELF) {
      struct ow_event e = { .type = OW_DETACH };
      watch_window_emit_event(&e);
      unwatch_window();
    }
    return;
  }
  if (event == EVENT_OBJECT_LOCATIONCHANGE) {
    if (hwnd == watch_target_info.hwnd && idObject == OBJID_WINDOW && idChild == CHILDID_SELF) {
      handle_movesize_event(&watch_target_info);
    }
    return;
  }
  if (event == EVENT_SYSTEM_FOREGROUND || event == EVENT_SYSTEM_MINIMIZEEND) {
    // checks if window is really gained focus
    // REASON: if multiple foreground windows switching too fast in short period,
    //         Windows sends EVENT_SYSTEM_FOREGROUND for them but MAY NOT actually
    //         focus window, so the focus is left on previous foreground window,
    //         but from the point of hook we think that focus is changed.
    if (GetForegroundWindow() == hwnd) {
      // printf("[1] EVENT_SYSTEM_FOREGROUND: OK, hwnd == GetForegroundWindow\n");
    } else {
      if (MSAA_check_window_focused_state(hwnd)) {
        // printf("[2] EVENT_SYSTEM_FOREGROUND: OK, GetForegroundWindow corrected by MSAA\n");
      } else {
        // printf("[2] EVENT_SYSTEM_FOREGROUND: FALSE POSITIVE\n");

        return;
      }
    }
    // check passed, continue normally

    handle_new_foreground(hwnd);
    return;
  }
}

static VOID CALLBACK foreground_timer_proc(HWND _hwnd, UINT msg, UINT_PTR timerId, DWORD dwmsEventTime)
{
  HWND system_foreground = GetForegroundWindow();
  if (
    foreground_window != system_foreground &&
    MSAA_check_window_focused_state(system_foreground)
  ) {
    handle_new_foreground(system_foreground);
  }
}

static void watch_thread_hook(void* _arg) {
  watch_target_info.foreground_hook = SetWinEventHook(
    EVENT_SYSTEM_FOREGROUND, EVENT_SYSTEM_FOREGROUND,
    NULL, hook_proc, 0, 0, WINEVENT_OUTOFCONTEXT);
  watch_target_info.minimized_hook = SetWinEventHook(
    EVENT_SYSTEM_MINIMIZEEND, EVENT_SYSTEM_MINIMIZEEND,
    NULL, hook_proc, 0, 0, WINEVENT_OUTOFCONTEXT);
  // FIXES: ForegroundLockTimeout (even when = 0); Also edge cases when apps stealing FG window.
  // NOTE:  Using timer because WH_SHELL & WH_CBT hooks require dll injection
  watch_target_info.timer_id = SetTimer(NULL, 0, FOREGROUND_TIMER_MS, foreground_timer_proc);

  DWORD pid;
  DWORD threadId = GetWindowThreadProcessId(watch_target_info.hwnd, &pid);
  if (threadId == 0) {
    return;
  }

  watch_target_info.location_hook = SetWinEventHook(
    EVENT_OBJECT_LOCATIONCHANGE, EVENT_OBJECT_LOCATIONCHANGE,
    NULL, hook_proc, 0, threadId,
    WINEVENT_OUTOFCONTEXT);
  watch_target_info.destroy_hook = SetWinEventHook(
    EVENT_OBJECT_DESTROY, EVENT_OBJECT_DESTROY,
    NULL, hook_proc, 0, threadId,
    WINEVENT_OUTOFCONTEXT);

  struct ow_event e = {
    .type = OW_ATTACH,
    .data.attach = {
      .has_access = -1,
      .is_fullscreen = -1
    }
  };
  e.data.attach.has_access = has_uipi_access(watch_target_info.hwnd);
  if (get_content_bounds(watch_target_info.hwnd, &e.data.attach.bounds)) {
    // emit OW_ATTACH
    watch_window_emit_event(&e);

    watch_target_info.is_focused = true;
    e.type = OW_FOCUS;
    watch_window_emit_event(&e);
  }
  else {
    // something went wrong, did the target window die right after becoming active?
    unwatch_window();
    return;
  }

  MSG message;
  while (GetMessageW(&message, (HWND)NULL, 0, 0) != FALSE) {
    if(watch_target_info.hwnd != NULL) {
      TranslateMessage(&message);
      DispatchMessageW(&message);
    }
  }
}

void unwatch_window() {
  if (watch_target_info.hwnd == NULL) {
    return;
  }

  UnhookWinEvent(watch_target_info.location_hook);
  UnhookWinEvent(watch_target_info.destroy_hook);
  UnhookWinEvent(watch_target_info.foreground_hook);
  UnhookWinEvent(watch_target_info.minimized_hook);
  KillTimer(NULL, watch_target_info.timer_id);
  watch_target_info.hwnd = NULL;
}

void watch_window(void* window_id) {
  unwatch_window();
  watch_target_info.hwnd = *((HWND*)window_id);
  uv_thread_create(&hook_tid, watch_thread_hook, NULL);
}

void set_parent(void* child_id, void* parent_id) {
  HWND child_hwnd = *((HWND*)child_id);
  LONG_PTR parent_hwnd = *((LONG_PTR*)parent_id);

  // SetParent(child_hwnd, parent_hwnd);
  SetWindowLongPtr(child_hwnd, -8, parent_hwnd); // set owner window
}

void reset_parent(void* child_id) {
  HWND child_hwnd = *((HWND*)child_id);
  SetWindowLongPtr(child_hwnd, -8, NULL); // reset owner window
}

void focus_window(void* window_id) {
  HWND hwnd = *((HWND*)window_id);
  SetForegroundWindow(hwnd);
}

int32_t get_session_id() {
  int32_t dwSessID = 0;
	ProcessIdToSessionId(GetCurrentProcessId(), &dwSessID);
  return dwSessID;
}
